from typing import (
    Any,
    Dict,
)

AUXILIARY_SCHEMA: Dict[str, Any] = {
    "enumerations": [
    ],
    "structures": [
        {
            "name": "Message",
            "properties": [
                {
                    "name": "jsonrpc",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                },
            ],
            "documentation": 'A general message as defined by JSON-RPC. The language server protocol\nalways uses “2.0” as the jsonrpc version.',
        },
        {
            "name": "RequestMessage",
            "properties": [
                {
                    "name": "id",
                    "type": {
                        "kind": "reference",
                        "name": "RequestId",
                    },
                    "documentation": "The request id.",
                },
                {
                    "name": "method",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                    "documentation": "The method to be invoked.",
                },
                {
                    "name": "params",
                    "type": {
                        "kind": "reference",
                        "name": "MessageParams",
                    },
                    "documentation": "The method's params.",
                    "optional": True,
                },
            ],
            "extends": [
                {
                    "kind": "reference",
                    "name": "Message",
                }
            ],
            "documentation": "A request message to describe a request between the client and the server.\nEvery processed request must send a response back to the sender of the\nrequest.",
        },
        {
            "name": "NotificationMessage",
            "properties": [
                {
                    "name": "method",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                    "documentation": "The method to be invoked.",
                },
                {
                    "name": "params",
                    "type": {
                        "kind": "reference",
                        "name": "MessageParams",
                    },
                    "documentation": "The notification's params.",
                    "optional": True,
                }
            ],
            "extends": [
                {
                    "kind": "reference",
                    "name": "Message",
                }
            ]
        },
        {
            "name": "ResponseError",
            "properties": [
                {
                    "name": "code",
                    "type": {
                        "kind": "base",
                        "name": "integer",
                    },
                    "documentation": "A number indicating the error type that occurred.",
                },
                {
                    "name": "message",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                    "documentation": "A string providing a short description of the error.",
                },
                {
                    "name": "data",
                    "type": {
                        "kind": "reference",
                        "name": "LSPAny",
                    },
                    "documentation": "A primitive or structured value that contains additional information about\nthe error. Can be omitted.",
                    "optional": True,
                },
            ],
        },
        {
            "name": "ResponseMessage",
            "properties": [
                {
                    "name": "id",
                    "type": {
                        "kind": "reference",
                        "name": "ResponseId",
                    },
                    "documentation": "The request id.",
                },
                {
                    "name": "result",
                    "type": {
                        "kind": "reference",
                        "name": "LSPAny",
                    },
                    "documentation": "The result of a request. This member is REQUIRED on success. This member\nMUST NOT exist if there was an error invoking the method.",
                    "optional": True,
                },
                {
                    "name": "error",
                    "type": {
                        "kind": "reference",
                        "name": "ResponseError",
                    },
                    "documentation": "The error object in case a request fails.",
                    "optional": True,
                },
            ],
            "extends": [
                {
                    "kind": "reference",
                    "name": "Message",
                },
            ],
            "documentation": "A Response Message sent as a result of a request. If a request doesn’t\nprovide a result value the receiver of a request still needs to return a\nresponse message to conform to the JSON-RPC specification. The result\nproperty of the ResponseMessage should be set to null in this case to signal\na successful request.",
        },
        {
            "name": "DocumentParams",
            "properties": [
                {
                    "name": "uri",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                },
            ],
        },
        {
            "name": "DocumentResult",
            "properties": [
                {
                    "name": "uri",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                },
                {
                    "name": "version",
                    "type": {
                        "kind": "base",
                        "name": "integer",
                    },
                },
                {
                    "name": "text",
                    "type": {
                        "kind": "base",
                        "name": "string",
                    },
                },
            ],
        },
    ],
    "typeAliases": [
        {
            "name": "RequestId",
            "type": {
                "kind": "or",
                "items": [
                    {
                        "kind": "base",
                        "name": "integer",
                    },
                    {
                        "kind": "base",
                        "name": "string",
                    },
                ],
            },
        },
        {
            "name": "MessageParams",
            "type": {
                "kind": "or",
                "items": [
                    {
                        "kind": "reference",
                        "name": "LSPArray",
                    },
                    {
                        "kind": "reference",
                        "name": "LSPObject",
                    },
                ],
            },
            "documentation": "A request message to describe a request between the client and the server.\nEvery processed request must send a response back to the sender of the\nrequest.",
        },
        {
            "name": "ResponseId",
            "type": {
                "kind": "or",
                "items": [
                    {
                        "kind": "base",
                        "name": "integer",
                    },
                    {
                        "kind": "base",
                        "name": "string",
                    },
                    {
                        "kind": "base",
                        "name": "null",
                    },
                ],
            },
        },
    ],
    "requests": [
		{
			"method": "$/document",
			"messageDirection": "clientToServer",
			"params": {
				"kind": "reference",
				"name": "DocumentParams",
			},
            "result": {
                "kind": "reference",
                "name": "DocumentResult",
            },
		},
        {
            "method": "$/telemetry",
            "messageDirection": "clientToServer",
            "result": {
                "kind": "reference",
				"name": "LSPAny"
            },
        },
    ],
    "notifications": [
    ],
}



from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List

import pytest

from lsprotocol.types import (CompletionItem, CompletionItemKind,
                              DidChangeConfigurationParams, DocumentHighlight,
                              DocumentSymbol, Hover, MarkupContent, MarkupKind,
                              Position, Range, SymbolKind)

from lfortran_language_server.lfortran_lsp_test_client import \
    LFortranLspTestClient

from llanguage_test_client.json_rpc import JsonArray
from llanguage_test_client.lsp_test_client import IncomingEvent


def test_goto_definition(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 18, 18
    doc.cursor = line, column
    doc.goto_definition()
    assert doc.line == 8
    assert doc.column == 5

def test_diagnostics(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 21, 1
    doc.cursor = line, column
    doc.write("error")
    validation = client.await_validation(doc.uri, doc.version)
    assert validation is not None
    assert validation["params"]["uri"] == doc.uri
    assert validation["params"]["version"] == doc.version
    diagnostics = validation["params"]["diagnostics"]
    assert len(diagnostics) == 2

    assert diagnostics[0]["message"] == "Statement or Declaration expected inside program, found Variable name"
    assert diagnostics[0]["range"]["start"]["line"] == 20
    assert diagnostics[0]["range"]["start"]["character"] == 0
    assert diagnostics[0]["range"]["end"]["line"] == 20
    assert diagnostics[0]["range"]["end"]["character"] == 5
    assert diagnostics[0]["severity"] == 1
    assert diagnostics[0]["source"] == "lfortran"

    assert diagnostics[1]["message"] == "Variable 'error' is not declared"
    assert diagnostics[1]["range"]["start"]["line"] == 20
    assert diagnostics[1]["range"]["start"]["character"] == 0
    assert diagnostics[1]["range"]["end"]["line"] == 20
    assert diagnostics[1]["range"]["end"]["character"] == 5
    assert diagnostics[1]["severity"] == 1
    assert diagnostics[1]["source"] == "lfortran"

    doc.backspace(5)
    validation = client.await_validation(doc.uri, doc.version)
    assert validation is not None
    assert validation["params"]["uri"] == doc.uri
    assert validation["params"]["version"] == doc.version
    diagnostics = validation["params"]["diagnostics"]
    assert len(diagnostics) == 0

def test_configuration_caching(client: LFortranLspTestClient) -> None:
    with NamedTemporaryFile(
            prefix="test_configuration_caching-",
            suffix=".f90",
            delete=True
    ) as tmp_file:
        doc = client.open_document("fortran", tmp_file.name)
        assert client.await_validation(doc.uri, doc.version) is not None
        doc.write("foo")
        doc.save()
        assert client.await_validation(doc.uri, doc.version) is not None

        def is_config_request_for_doc(event: IncomingEvent) -> bool:
            return event.data.get("method", None) == "workspace/configuration" and \
                len(event.data["params"]["items"]) > 0 and \
                event.data["params"]["items"][0]["scopeUri"] == doc.uri and \
                event.data["params"]["items"][0]["section"] == "LFortran"

        event, index = client.find_incoming_event(is_config_request_for_doc)
        # The config must be requested before it may be cached:
        assert event is not None

        doc.write("bar")
        assert client.await_validation(doc.uri, doc.version) is not None
        event, index = client.find_incoming_event(is_config_request_for_doc, index + 1)
        # The config has been cached, so there is no need to cache it again:
        assert event is None

        # Invalidate the document caches:
        client.send_workspace_did_change_configuration(
            DidChangeConfigurationParams(client.config)
        )

        doc.write("baz")
        assert client.await_validation(doc.uri, doc.version) is not None
        event, index = client.find_incoming_event(is_config_request_for_doc, index + 1)
        # The cache was invalidated so the config must be requested again:
        assert event is not None

def test_rename(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 4, 20
    doc.cursor = line, column
    doc.rename("foo")
    assert doc.text == "\n".join([
        "module module_function_call1",
        "    type :: softmax",
        "    contains",
        "      foo",
        "    end type softmax",
        "  contains",
        "  ",
        "    pure function foo(self, x) result(res)",
        "      class(softmax), intent(in) :: self",
        "      real, intent(in) :: x(:)",
        "      real :: res(size(x))",
        "    end function foo",
        "  ",
        "    pure function eval_1d_prime(self, x) result(res)",
        "      class(softmax), intent(in) :: self",
        "      real, intent(in) :: x(:)",
        "      real :: res(size(x))",
        # TODO: Once lfortran can rename the function call, swap the following
        # two lines:
        # "      res = self%foo(x)",
        "      res = self%eval_1d(x)",
        "    end function eval_1d_prime",
        "end module module_function_call1",
    ]) + "\n"

def test_document_highlight(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent.parent / "examples" / "expr2.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 7, 10
    doc.cursor = line, column
    doc.highlight_symbol()
    assert doc.symbol_highlights is not None
    # NOTE: DocumentHighlight is not hashable, so we cannot perform set comparison ...
    expected_highlights: List[DocumentHighlight] = [
        DocumentHighlight(
            range=Range(
                start=Position(
                    line=3,
                    character=11,
                ),
                end=Position(
                    line=3,
                    character=12,
                ),
            ),
        ),
        DocumentHighlight(
            range=Range(
                start=Position(
                    line=5,
                    character=0,
                ),
                end=Position(
                    line=5,
                    character=1,
                ),
            ),
        ),
        DocumentHighlight(
            range=Range(
                start=Position(
                    line=6,
                    character=9,
                ),
                end=Position(
                    line=6,
                    character=10,
                ),
            ),
        ),
    ]
    for highlight in doc.symbol_highlights:
        expected_highlights.remove(highlight)
    assert len(expected_highlights) == 0


def test_document_hover(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 18, 22
    doc.cursor = line, column
    doc.hover()
    assert doc.preview == Hover(
        contents=MarkupContent(
            kind=MarkupKind.Markdown,
            value="```fortran\nfunction eval_1d(self, x) result(res)\n    class(softmax), intent(in) :: self\n    real(4)[:], intent(in) :: x\n    real(4)[:] :: res\nend function\n```"
        ),
        range=Range(
            end=Position(
                character=24,
                line=11
            ),
            start=Position(
                character=4,
                line=7
            )
        )
    )


def test_symbol_tree(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent.parent / "examples" / "expr2.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    assert doc.symbols == [
        DocumentSymbol(
            children=[
                DocumentSymbol(
                    kind=SymbolKind.Variable,
                    name="x",
                    range=Range(
                        start=Position(
                            line=3,
                            character=11,
                        ),
                        end=Position(
                            line=3,
                            character=12,
                        ),
                    ),
                    selection_range=Range(
                        start=Position(
                            line=3,
                            character=11,
                        ),
                        end=Position(
                            line=3,
                            character=12,
                        ),
                    ),
                ),
            ],
            kind=SymbolKind.Function,
            name="expr2",
            range=Range(
                start=Position(
                    line=0,
                    character=0,
                ),
                end=Position(
                    line=8,
                    character=11,
                ),
            ),
            selection_range=Range(
                start=Position(
                    line=0,
                    character=0,
                ),
                end=Position(
                    line=8,
                    character=11,
                ),
            ),
        ),
    ]
    line, column = 4, 12
    doc.cursor = line, column
    doc.rename("y")  # x -> y
    assert doc.symbols == [
        DocumentSymbol(
            children=[
                DocumentSymbol(
                    kind=SymbolKind.Variable,
                    name="y",
                    range=Range(
                        start=Position(
                            line=3,
                            character=11,
                        ),
                        end=Position(
                            line=3,
                            character=12,
                        ),
                    ),
                    selection_range=Range(
                        start=Position(
                            line=3,
                            character=11,
                        ),
                        end=Position(
                            line=3,
                            character=12,
                        ),
                    ),
                ),
            ],
            kind=SymbolKind.Function,
            name="expr2",
            range=Range(
                start=Position(
                    line=0,
                    character=0,
                ),
                end=Position(
                    line=8,
                    character=11,
                ),
            ),
            selection_range=Range(
                start=Position(
                    line=0,
                    character=0,
                ),
                end=Position(
                    line=8,
                    character=11,
                ),
            ),
        ),
    ]

def test_semantic_highlighting(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    doc.semantic_highlight()
    expected_highlights = [
        0,0,6,15,0, # Keyword at line=0, column=0, length=6, text=`module`
        0,7,21,8,0, # Variable at line=0, column=7, length=21, text=`module_function_call1`
        1,4,4,15,0, # Keyword at line=1, column=4, length=4, text=`type`
        0,8,7,8,0,  # Variable at line=1, column=12, length=7, text=`softmax`
        1,4,8,15,0, # Keyword at line=2, column=4, length=8, text=`contains`
        1,6,9,8,0,  # Variable at line=3, column=6, length=9, text=`procedure`
        0,13,7,8,0, # Variable at line=3, column=19, length=7, text=`eval_1d`
        1,4,3,15,0, # Keyword at line=4, column=4, length=3, text=`end`
        0,4,4,15,0, # Keyword at line=4, column=8, length=4, text=`type`
        0,5,7,8,0,  # Variable at line=4, column=13, length=7, text=`softmax`
        1,2,8,15,0, # Keyword at line=5, column=2, length=8, text=`contains`
        2,4,4,8,0,  # Variable at line=7, column=4, length=4, text=`pure`
        0,5,8,15,0, # Keyword at line=7, column=9, length=8, text=`function`
        0,9,7,8,0,  # Variable at line=7, column=18, length=7, text=`eval_1d`
        0,8,4,8,0,  # Variable at line=7, column=26, length=4, text=`self`
        0,6,1,8,0,  # Variable at line=7, column=32, length=1, text=`x`
        0,3,6,8,0,  # Variable at line=7, column=35, length=6, text=`result`
        0,7,3,8,0,  # Variable at line=7, column=42, length=3, text=`res`
        1,6,5,8,0,  # Variable at line=8, column=6, length=5, text=`class`
        0,6,7,8,0,  # Variable at line=8, column=12, length=7, text=`softmax`
        0,10,6,8,0, # Variable at line=8, column=22, length=6, text=`intent`
        0,7,2,8,0,  # Variable at line=8, column=29, length=2, text=`in`
        0,7,4,8,0,  # Variable at line=8, column=36, length=4, text=`self`
        1,6,4,15,0, # Keyword at line=9, column=6, length=4, text=`real`
        0,6,6,8,0,  # Variable at line=9, column=12, length=6, text=`intent`
        0,7,2,8,0,  # Variable at line=9, column=19, length=2, text=`in`
        0,7,1,8,0,  # Variable at line=9, column=26, length=1, text=`x`
        1,6,4,15,0, # Keyword at line=10, column=6, length=4, text=`real`
        0,8,3,8,0,  # Variable at line=10, column=14, length=3, text=`res`
        0,4,4,8,0,  # Variable at line=10, column=18, length=4, text=`size`
        0,5,1,8,0,  # Variable at line=10, column=23, length=1, text=`x`
        1,4,3,15,0, # Keyword at line=11, column=4, length=3, text=`end`
        0,4,8,15,0, # Keyword at line=11, column=8, length=8, text=`function`
        0,9,7,8,0,  # Variable at line=11, column=17, length=7, text=`eval_1d`
        2,4,4,8,0,  # Variable at line=13, column=4, length=4, text=`pure`
        0,5,8,15,0, # Keyword at line=13, column=9, length=8, text=`function`
        0,9,13,8,0, # Variable at line=13, column=18, length=13, text=`eval_1d_prime`
        0,14,4,8,0, # Variable at line=13, column=32, length=4, text=`self`
        0,6,1,8,0,  # Variable at line=13, column=38, length=1, text=`x`
        0,3,6,8,0,  # Variable at line=13, column=41, length=6, text=`result`
        0,7,3,8,0,  # Variable at line=13, column=48, length=3, text=`res`
        1,6,5,8,0,  # Variable at line=14, column=6, length=5, text=`class`
        0,6,7,8,0,  # Variable at line=14, column=12, length=7, text=`softmax`
        0,10,6,8,0, # Variable at line=14, column=22, length=6, text=`intent`
        0,7,2,8,0,  # Variable at line=14, column=29, length=2, text=`in`
        0,7,4,8,0,  # Variable at line=14, column=36, length=4, text=`self`
        1,6,4,15,0, # Keyword at line=15, column=6, length=4, text=`real`
        0,6,6,8,0,  # Variable at line=15, column=12, length=6, text=`intent`
        0,7,2,8,0,  # Variable at line=15, column=19, length=2, text=`in`
        0,7,1,8,0,  # Variable at line=15, column=26, length=1, text=`x`
        1,6,4,15,0, # Keyword at line=16, column=6, length=4, text=`real`
        0,8,3,8,0,  # Variable at line=16, column=14, length=3, text=`res`
        0,4,4,8,0,  # Variable at line=16, column=18, length=4, text=`size`
        0,5,1,8,0,  # Variable at line=16, column=23, length=1, text=`x`
        1,6,3,8,0,  # Variable at line=17, column=6, length=3, text=`res`
        0,6,4,8,0,  # Variable at line=17, column=12, length=4, text=`self`
        0,5,7,8,0,  # Variable at line=17, column=17, length=7, text=`eval_1d`
        0,8,1,8,0,  # Variable at line=17, column=25, length=1, text=`x`
        1,4,3,15,0, # Keyword at line=18, column=4, length=3, text=`end`
        0,4,8,15,0, # Keyword at line=18, column=8, length=8, text=`function`
        0,9,13,8,0, # Variable at line=18, column=17, length=13, text=`eval_1d_prime`
        1,0,3,15,0, # Keyword at line=19, column=0, length=3, text=`end`
        0,4,6,15,0, # Keyword at line=19, column=4, length=6, text=`module`
        0,7,21,8,0, # Variable at line=19, column=11, length=21, text=`module_function_call1`
    ]
    assert doc.semantic_highlights is not None
    assert doc.semantic_highlights.data == expected_highlights

def test_code_completion(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    line, column = 18, -1
    doc.cursor = line, column
    doc.newline()
    doc.write("self")
    doc.complete()
    expected_completions = [
        CompletionItem(
            label="self",
            kind=CompletionItemKind.Variable
        )
    ]
    assert doc.completions is not None
    assert isinstance(doc.completions, list)
    for completion in doc.completions:
        expected_completions.remove(completion)
    assert len(expected_completions) == 0
    doc.write("%")
    doc.complete()
    expected_completions = [
        CompletionItem(
            label="module_function_call1",
            kind=CompletionItemKind.Module,
        ),
        CompletionItem(
            label="eval_1d",
            kind=CompletionItemKind.Function,
        ),
        CompletionItem(
            label="res",
            kind=CompletionItemKind.Variable,
        ),
        CompletionItem(
            label="self",
            kind=CompletionItemKind.Variable,
        ),
        CompletionItem(
            label="x",
            kind=CompletionItemKind.Variable,
        ),
        CompletionItem(
            label="eval_1d_prime",
            kind=CompletionItemKind.Function,
        ),
        CompletionItem(
            label="1_softmax_eval_1d",
            kind=CompletionItemKind.Function,
        ),
        CompletionItem(
            label="softmax",
            kind=CompletionItemKind.Struct,
        ),
    ]
    assert doc.completions is not None
    assert isinstance(doc.completions, list)
    for completion in doc.completions:
        expected_completions.remove(completion)
    assert len(expected_completions) == 0

def test_whole_document_formatting(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    doc.format()
    assert doc.text == "\n".join([
        "module module_function_call1",
        "    type :: softmax",
        "",
        "contains",
        "",
        "procedure :: eval_1d",
        "    end type softmax",
        "",
        "contains",
        "",
        "    pure function eval_1d(self, x) result(res)",
        "        class(softmax), intent(in) :: self",
        "        real, intent(in) :: x(:)",
        "        real :: res(size(x))",
        "    end function eval_1d",
        "",
        "",
        "    pure function eval_1d_prime(self, x) result(res)",
        "        class(softmax), intent(in) :: self",
        "        real, intent(in) :: x(:)",
        "        real :: res(size(x))",
        "        res = self%eval_1d(x)",
        "    end function eval_1d_prime",
        "",
        "end module module_function_call1",
        ""
    ])

def test_partial_document_formatting(client: LFortranLspTestClient) -> None:
    path = Path(__file__).absolute().parent.parent.parent / "function_call1.f90"
    doc = client.open_document("fortran", path)
    assert client.await_validation(doc.uri, doc.version) is not None
    doc.select(14, 5, 19, 31)
    doc.format_range()
    assert doc.text == "\n".join([
        "module module_function_call1",
        "    type :: softmax",
        "    contains",
        "      procedure :: eval_1d",
        "    end type softmax",
        "  contains",
        "  ",
        "    pure function eval_1d(self, x) result(res)",
        "      class(softmax), intent(in) :: self",
        "      real, intent(in) :: x(:)",
        "      real :: res(size(x))",
        "    end function eval_1d",
        "  ",
        "    pure function eval_1d_prime(self, x) result(res)",
        "        class(softmax), intent(in) :: self",
        "        real, intent(in) :: x(:)",
        "        real :: res(size(x))",
        "        res = self%eval_1d(x)",
        "    end function eval_1d_prime",
        "",
        "end module module_function_call1",
        ""
    ])

def test_telemetry_event(client: LFortranLspTestClient) -> None:
    telemetry: JsonArray = client.get_telemetry()
    if not any(filter(lambda event: event["key"] == "processUsage", telemetry)):
        pytest.skip("ProcessUsage is not supported on this platform")



#!/usr/bin/env python3

import argparse
import os
import sys
import traceback
from pathlib import Path

from llanguage_server.cxx.lsp_code_generator import CPlusPlusLspCodeGenerator

EXIT_SUCCESS: int = 0
EXIT_FAILURE: int = 1

def is_debug_enabled() -> bool:
    debug: str = os.environ.get("DEBUG_LSP_CODE_GENERATOR", "false")
    return debug.lower() in ["true", "t", "yes", "y", "on", "1"]

def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="LSPCodeGenerator",
        description="Generates code in a target language from Microsoft's LSP spec (Language Server Protocol).",
    )
    parser.add_argument(
        "-s", "--schema",
        help="Path to the metaModel.json file describing the current LSP spec (e.g. version 3.17).",
        type=argparse.FileType(mode="r", encoding="UTF-8"),
    )
    parser.add_argument(
        "-l", "--target-language",
        help="Target language for the generated code.",
        choices=["c++"],
    )
    parser.add_argument(
        "-o", "--output-dir",
        help="Path to the directory into which to write the generated files.",
        type=Path,
    )
    return parser.parse_args()

def main() -> int:
    args: argparse.Namespace = parse_args()
    match args.target_language:
        case "c++":
            code_generator = CPlusPlusLspCodeGenerator(args)
        case _:
            raise ValueError(f"Unsupported target language: {args.target_language}")
    code_generator.generate_code()
    return EXIT_SUCCESS

if __name__ == "__main__":
    try:
        exit_code = main()
        sys.exit(exit_code)
    except SystemExit as system_exit:
        if system_exit.code is not None:
            if system_exit.code != 0 and is_debug_enabled():
                traceback.print_exc()
            sys.exit(system_exit.code)
    except:
        print("Rescued unhandled exception:", file=sys.stderr)
        traceback.print_exc()
        sys.exit(EXIT_FAILURE)



from abc import abstractmethod
from pathlib import Path
from typing import Dict, Tuple

from llanguage_server.cxx.lsp_file_generator import CPlusPlusLspFileGenerator
from llanguage_server.lsp.datatypes import LspSpec, LspSymbol
from llanguage_server.lsp.visitors import (BaseLspSymbolVisitor,
                                           LspAnalysisPipeline)


class BaseCPlusPlusLspVisitor(CPlusPlusLspFileGenerator, BaseLspSymbolVisitor):

    def __init__(
            self,
            file_path: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        CPlusPlusLspFileGenerator.__init__(self, file_path, schema, pipeline, namespace)
        BaseLspSymbolVisitor.__init__(self, pipeline.indexer)

    def generate_structure(self, struct_symbol: LspSymbol) -> None:
        pass

    def generate_union(self, union_symbol: LspSymbol) -> None:
        pass

    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> None:
        pass

    def visit_array_symbol(self, array_symbol: LspSymbol) -> None:
        pass

    def visit_base_symbol(self, base_symbol: LspSymbol) -> None:
        pass

    def visit_enumeration_symbol(self, enum_symbol: LspSymbol) -> None:
        with self.nested_name_as(enum_symbol.name):
            enum_spec = enum_symbol.spec
            self.generate_enumeration(enum_spec)

    def visit_inner_symbol(self, inner_symbol: LspSymbol) -> None:
        with self.nested_name_as(inner_symbol.name):
            self.generate_structure(inner_symbol)

    def visit_map_symbol(self, map_symbol: LspSymbol) -> None:
        pass

    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        pass

    def visit_reference_symbol(self, reference_symbol: LspSymbol) -> None:
        pass

    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        pass

    def visit_structure_symbol(self, struct_symbol: LspSymbol) -> None:
        with self.nested_name_as(struct_symbol.name):
            self.generate_structure(struct_symbol)

    def visit_tuple_symbol(self, tuple_symbol: LspSymbol) -> None:
        pass

    def visit_union_symbol(self, union_symbol: LspSymbol) -> None:
        with self.nested_name_as(union_symbol.name):
            self.generate_union(union_symbol)

    def generate_code(self) -> None:
        for symbol in self.pipeline.indexer.topological_sort():
            self.visit_symbol(symbol)



import sys
import os

intrinsic_funcs_args = {
    "Kind": [
        {
            "args": [("int",), ("real",), ("bool",), ("char",), (("complex",))],
            "return": "int32"
        },
    ],
    "FMA": [
        {
            "args": [("real", "real", "real")],
            "ret_type_arg_idx": 0
        }
    ],
    "FlipSign": [
        {
            "args": [("int", "real")],
            "ret_type_arg_idx": 1
        }
    ],
    "FloorDiv": [
        {
            "args": [("int", "int"), ("uint", "uint"), ("real", "real"), ("bool", "bool")],
            "ret_type_arg_idx": 0
        },
    ],
    "Mod": [
        {
            "args": [("int", "int"), ("real", "real")],
            "ret_type_arg_idx": "dynamic"
        },
    ],
    "Trailz": [
        {
            "args": [("int",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Spacing": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Modulo": [
        {
            "args": [("int", "int"), ("real", "real")],
            "ret_type_arg_idx": 0
        },
    ],
    "BesselJN": [
        {
            "args": [("int", "real")],
            "ret_type_arg_idx": 1
        },
    ],
    "BesselYN": [
        {
            "args": [("int", "real")],
            "ret_type_arg_idx": 1
        },
    ],
    "Leadz": [
        {
            "args": [("int",)],
            "ret_type_arg_idx": 0
        },
    ],
    "ToLowerCase": [
        {
            "args": [("char",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Hypot": [
        {
            "args": [("real", "real")],
            "ret_type_arg_idx": 0,
            "same_kind_arg" : 2
        }
    ],
    "Trunc": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Gamma": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "LogGamma": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Log10": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Erf": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Erfc": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Exp": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "ErfcScaled": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Atan2": [
        {
            "args": [("real", "real")],
            "ret_type_arg_idx": 0
        }
    ],
    "Atan2d": [
        {
            "args": [("real", "real")],
            "ret_type_arg_idx": 0
        }
    ],
    "Fix": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Exp2": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Expm1": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "SelectedIntKind": [
        {
            "args": [("int",)],
            "return": "int32"
        }
    ],
    "SelectedRealKind": [
        {
            "args": [("int", "int", "int")],
            "return": "int32"
        }
    ],
    "SelectedCharKind": [
        {
            "args": [("char",)],
            "return": "int32"
        }
    ],
    "SelectedLogicalKind": [
        {
            "args": [("int",)],
            "return": "int32"
        }
    ],
    "Logical": [
        {
            "args": [("bool", )],
            "return": "logical",
            "kind_arg": True
        }
    ],
    "Digits": [
        {
            "args": [("int",), ("real",)],
            "return": "int32"
        },
    ],
    "StringContainsSet": [
        {
            "args": [("char", "char", "bool", "int")],
            "ret_type_arg_idx": 3
        }
    ],
    "StringFindSet": [
        {
            "args": [("char", "char", "bool", "int")],
            "ret_type_arg_idx": 3
        }
    ],
    "SubstrIndex": [
        {
            "args": [("char", "char", "bool", "int")],
            "ret_type_arg_idx": 3
        }
    ],
    "MinExponent": [
        {
            "args": [("real",)],
            "return": "int32"
        }
    ],
    "MaxExponent": [
        {
            "args": [("real",)],
            "return": "int32"
        }
    ],
    "Partition": [
        {
            "args": [("char", "char")],
            "ret_type_arg_idx": 0
        }
    ],
    "ListReverse": [
        {
            "args": [("list",)],
            "return": "nullptr"
        }
    ],
    "ListReserve": [
        {
            "args": [("list", "int")],
            "return": "nullptr"
        }
    ],
    "Sign": [
        {
            "args": [("int", "int"), ("real", "real")],
            "ret_type_arg_idx": 0,
            "same_kind_arg": 2
        },
    ],
    "Radix": [
        {
            "args": [("int",), ("real",)],
            "return": "int32"
        },
    ],
    "OutOfRange": [
        {
            "args": [("int", "real", "bool"), ("real", "real", "bool"), ("int", "int", "bool"), ("real", "int", "bool")],
            "return": "logical"
        },
    ],
    "StorageSize": [
        {
            "args": [("any",)],
            "return": "int32",
            "kind_arg": True
        },
    ],
    "Nearest": [
        {
            "args": [("real", "real")],
            "ret_type_arg_idx": 0
        },
    ],
    "Adjustl": [
        {
            "args": [("char",)],
            "ret_type_arg_idx": 0,
            "allow_polymorphic_arg": [0]
        }
    ],
    "Adjustr": [
        {
            "args": [("char",)],
            "ret_type_arg_idx": 0,
            "allow_polymorphic_arg": [0]
        }
    ],
    "Aint": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0,
            "kind_arg": True
        }
    ],
    "Isnan": [
        {
            "args": [("real",)],
            "return": "logical",
        }
    ],
    "SameTypeAs": [
        {
            "args": [("any", "any")],
            "return": "logical"
        }
    ],
    "ExtendsTypeOf": [
        {
            "args": [("any", "any")],
            "return": "logical"
        }
    ],
    "Nint": [
        {
            "args": [("real",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "Idnint": [
        {
            "args": [("real",)],
            "return": "int32",
            "kind_validation": [{"first":{0: 8}}]
        }
    ],
    "Anint": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0,
            "kind_arg": True
        }
    ],
    "Floor": [
        {
            "args": [("real",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "Ceiling": [
        {
            "args": [("real",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "Asind": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Acosd": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Atand": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Sind": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Cosd": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Tand": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Sinpi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Cospi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Tanpi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Asinpi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Acospi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Atanpi": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Atan2pi": [
        {
            "args": [("real", "real")],
            "ret_type_arg_idx": 0
        }
    ],
    "BesselJ0": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "BesselJ1": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "BesselY0": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "BesselY1": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Sqrt": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Sin": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Cos": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Tan": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Asin": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Acos": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Atan": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Sinh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Cosh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Tanh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Asinh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Acosh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Atanh": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Log": [
        {
            "args": [("real",), ("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Sngl": [
        {
            "args": [("real",)],
            "return": "real32",
        }
    ],
    "SignFromValue": [
        {
            "args": [("int", "int"), ("real", "real")],
            "ret_type_arg_idx": 0
        },
    ],
    "Ifix": [
        {
            "args": [("real",)],
            "return": "int32",
            "kind_validation": [{"first":{0: 4}}]
        }
    ],
    "Idint": [
        {
            "args": [("real",)],
            "kind_validation": [{"first":{0: 8}}],
            "return": "int32"
        }
    ],
    "Ishft": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Bgt": [
        {
            "args": [("int", "int")],
            "return": "logical"
        },
    ],
    "Blt": [
        {
            "args": [("int", "int")],
            "return": "logical"
        },
    ],
    "Bge": [
        {
            "args": [("int", "int")],
            "return": "logical"
        },
    ],
    "Ble": [
        {
            "args": [("int", "int")],
            "return": "logical"
        },
    ],
    "Lgt": [
        {
            "args": [("char", "char")],
            "return": "logical"
        },
    ],
    "Llt": [
        {
            "args": [("char", "char")],
            "return": "logical"
        },
    ],
    "Lge": [
        {
            "args": [("char", "char")],
            "return": "logical"
        },
    ],
    "Lle": [
        {
            "args": [("char", "char")],
            "return": "logical"
        },
    ],
    "Not": [
        {
            "args": [("int",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Iand": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0,
            "same_kind_arg": 2
        },
    ],
    "And": [
        {
            "args": [("int", "int"),("bool","bool")],
            "ret_type_arg_idx": 0,
        },
    ],
    "Ior": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0,
            "same_kind_arg": 2
        },
    ],
    "Or": [
        {
            "args": [("int", "int"), ("bool", "bool")],
            "ret_type_arg_idx": 0,
        },
    ],
    "Ieor": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0,
            "same_kind_arg": 2
        },
    ],
    "Xor": [
        {
            "args": [("int", "int"), ("bool", "bool")],
            "ret_type_arg_idx": 0,
        },
    ],
    "Ibclr": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Ibset": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Btest": [
        {
            "args": [("int", "int")],
            "return": "logical"
        },
    ],
    "Ibits": [
        {
            "args": [("int", "int", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Shiftr": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        }
    ],
    "Rshift": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        }
    ],
    "Shiftl": [
        {
            "args": [("int", "int")],
            "ret_type_arg_idx": 0
        }
    ],
    "Aimag": [
        {
            "args": [("complex",)],
            "return": "real32",
            "kind_arg": True
        },
    ],
    "Dreal": [
        {
            "args": [("complex64",)],
            "return": "real64",
        },
    ],
    "BitSize": [
        {
            "args": [("int",)],
            "ret_type_arg_idx": 0
        }
    ],
    "NewLine": [
        {
            "args": [("char",)],
            "return": "character(1)"
        }
    ],
    "Range": [
        {
            "args": [("int",), ("real",), ("complex",)],
            "return": "int32"
        },
    ],
    "Epsilon": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Precision": [
        {
            "args": [("real",), ("complex",)],
            "return": "int32"
        }
    ],
    "Tiny": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Conjg": [
        {
            "args": [("complex",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Scale": [
        {
            "args": [("real", "int")],
            "ret_type_arg_idx": 0
        }
    ],
    "Huge": [
        {
            "args": [("int",), ("real",)],
            "ret_type_arg_idx": 0
        }
    ],
    "Loc": [
        {
            "args": [("any",)],
            "return": "int32"
        }
    ],
    "Dprod": [
        {
            "args": [("real", "real")],
            "kind_validation": [{"first":{0: 4}}, {"second":{1: 4}}],
            "return": "real64"
        }
    ],
    "Dim": [
        {
            "args": [("int", "int"), ("real", "real")],
            "ret_type_arg_idx": 0
        },
    ],
    "Maskl": [
        {
            "args": [("int",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "Maskr": [
        {
            "args": [("int",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "Merge": [
        {
            "args": [("any", "any", "bool")],
            "ret_type_arg_idx": 0
        }
    ],
    "Mergebits": [
        {
            "args": [("int", "int", "int")],
            "ret_type_arg_idx": 0,
            "same_kind_arg": 3
        }
    ],
    "Ishftc": [
        {
            "args": [("int", "int", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Ichar": [
        {
            "args": [("char",)],
            "return": "int32",
            "kind_arg": True,
            "char_len_validation": [
                {1: {0: {"min": 1, "max": 1}}}
            ]
        },
    ],
    "Char": [
        {
            "args": [("int",)],
            "return": "character(1)",
            "kind_arg": True
        }
    ],
    "Achar": [
        {
            "args": [("int",)],
            "return": "character(1)",
            "kind_arg": True
        }
    ],
    "Iachar": [
        {
            "args": [("char",)],
            "return": "int32",
            "kind_arg": True,
            "char_len_validation": [
                {1: {0: {"min": 1, "max": 1}}}
            ]
        },
    ],
    "Exponent": [
        {
            "args": [("real",)],
            "return": "int32",
        },
    ],
    "Fraction": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "SetExponent": [
        {
            "args": [("real", "int")],
            "ret_type_arg_idx": 0
        },
    ],
    "Rrspacing": [
        {
            "args": [("real",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Dshiftl": [
       {
           "args": [("int", "int", "int",)],
           "ret_type_arg_idx": 0,
           "same_kind_arg": 2
       },
    ],
    "Dshiftr": [
        {
            "args": [("int", "int", "int",)],
            "ret_type_arg_idx": 0
        },
    ],
    "Popcnt": [
        {
            "args": [("int",)],
            "return": "int32",
        },
    ],
    "Poppar": [
        {
            "args": [("int",)],
            "return": "int32",
        },
    ],
    "Real": [
        {
            "args": [("int",), ("real",), ("complex",)],
            "return": "real32",
            "kind_arg": True,
            "real_32_except_complex": True
        },
    ],
    "Int": [
        {
            "args": [("int",), ("real",), ("complex",)],
            "return": "int32",
            "kind_arg": True
        }
    ],
    "StringLenTrim": [
        {
            "args": [("char",)],
            "return": "int32",
            "kind_arg": True,
            "allow_polymorphic_arg": [0]
        }
    ],
    "StringTrim": [
        {
            "args": [("char",)],
            "return" : "allocatable_deferred_string()",
            "allow_polymorphic_arg": [0]
        }
    ],
}

skip_create_func = ["Partition"]
compile_time_only_fn = [
    "Epsilon",
    "Radix",
    "IsContiguous",
    "StorageSize",
    "Range",
    "Precision",
    "Tiny",
    "Huge",
    "Loc",
    "BitSize",
    "NewLine",
    "Kind",
    "MaxExponent",
    "MinExponent",
    "SameTypeAs",
    "ExtendsTypeOf",
    "Digits",
]

type_to_asr_type_check = {
    "any": "!ASR::is_a<ASR::TypeParameter_t>",
    "int": "is_integer",
    "uint": "is_unsigned_integer",
    "real": "is_real",
    "bool": "is_logical",
    "char": "is_character",
    "complex": "is_complex",
    "complex64": "is_complex<8>",
    "dict": "ASR::is_a<ASR::Dict_t>",
    "list": "ASR::is_a<ASR::List_t>",
    "tuple": "ASR::is_a<ASR::Tuple_t>"
}

intrinsic_funcs_ret_type = {
    "Kind": ["int"],
    "Partition": ["tuple"],
    "ListReverse": ["null"],
    "ListReserve": [ "null"],
    "Radix": ["int"],
}

src = ""
indent = "    "

def compute_arg_types(indent, no_of_args, args_arr):
    global src
    for i in range(no_of_args):
        src += indent + f"ASR::ttype_t *arg_type{i} = ASRUtils::expr_type({args_arr}[{i}]);\n"

def compute_arg_kinds(indent, no_of_args):
    global src
    for i in range(no_of_args):
        src += indent + f"int kind{i} = ASRUtils::extract_kind_from_ttype_t(arg_type{i});\n"

def compute_arg_condition(no_of_args, args_lists, allow_polymorphic_arg, args_var):
    condition = []
    cond_in_msg = []
    for arg_list in args_lists:
        subcond = []
        subcond_in_msg = []
        for i in range(no_of_args):
            arg = arg_list[i]
            type_check = f"{type_to_asr_type_check[arg]}(*arg_type{i})"
            # Add unlimited polymorphic check if specified
            if allow_polymorphic_arg and i in allow_polymorphic_arg:
                type_check = f"({type_check} || (is_unlimited_polymorphic_type({args_var}[{i}])))"
            subcond.append(type_check)
            subcond_in_msg.append(arg)
        condition.append(" && ".join(subcond))
        cond_in_msg.append(", ".join(subcond_in_msg))
    return (f"({') || ('.join(condition)})", f"({') or ('.join(cond_in_msg)})")

def compute_kind_condition(no_of_args):
    condition = []
    for i in range(1, no_of_args):
        condition.append(f"kind0 == kind{i}")
    return f"({') && ('.join(condition)})"

def add_verify_arg_type_src(func_name):
    global src
    arg_infos = intrinsic_funcs_args[func_name]
    same_kind_arg = arg_infos[0].get("same_kind_arg", False)
    no_of_args_msg = ""
    for i, arg_info in enumerate(arg_infos):
        args_lists = arg_info["args"]
        no_of_args = len(args_lists[0])
        no_of_args_msg += " or " if i > 0 else ""
        no_of_args_msg += f"{no_of_args}"
        else_if = "else if" if i > 0 else "if"
        src += 2 * indent + f"{else_if} (x.n_args == {no_of_args}) " + " {\n"
        src += 3 * indent + f'ASRUtils::require_impl(x.m_overload_id == {i}, "Overload Id for {func_name} expected to be {i}, found " + std::to_string(x.m_overload_id), x.base.base.loc, diagnostics);\n'
        compute_arg_types(3 * indent, no_of_args, "x.m_args")
        allow_polymorphic_arg = arg_info.get("allow_polymorphic_arg", None)
        condition, cond_in_msg = compute_arg_condition(no_of_args, args_lists, allow_polymorphic_arg, "x.m_args")
        src += 3 * indent + f'ASRUtils::require_impl({condition}, "Unexpected args, {func_name} expects {cond_in_msg} as arguments", x.base.base.loc, diagnostics);\n'
        if same_kind_arg:
            compute_arg_kinds(3 * indent, same_kind_arg)
            condition = compute_kind_condition(same_kind_arg)
            src += 3 * indent + f'ASRUtils::require_impl({condition}, "Kind of all the arguments of {func_name} must be the same", x.base.base.loc, diagnostics);\n'
        src += 2 * indent + "}\n"
    src += 2 * indent + "else {\n"
    src += 3 * indent + f'ASRUtils::require_impl(false, "Unexpected number of args, {func_name} takes {no_of_args_msg} arguments, found " + std::to_string(x.n_args), x.base.base.loc, diagnostics);\n'
    src += 2 * indent + "}\n"

def add_verify_return_type_src(func_name):
    if func_name not in intrinsic_funcs_ret_type.keys():
        return ""
    global src
    ret_type_cond = ""
    ret_type_cond_in_msg = ""
    for i, ret_type in enumerate(intrinsic_funcs_ret_type[func_name]):
        if ret_type == "null":
            ret_type_cond += f"x.m_type == nullptr"
        else:
            ret_type_cond += f"{type_to_asr_type_check[ret_type]}(*x.m_type)"
        ret_type_cond_in_msg += f"{ret_type}"
        if i < len(intrinsic_funcs_ret_type[func_name]) - 1:
            ret_type_cond += " || "
            ret_type_cond_in_msg += " or "
    src += 2 * indent + f'ASRUtils::require_impl({ret_type_cond}, "Unexpected return type, {func_name} expects `{ret_type_cond_in_msg}` as return type", x.base.base.loc, diagnostics);\n'

def add_create_func_arg_type_src(func_name):
    global src
    arg_infos = intrinsic_funcs_args[func_name]
    no_of_args_msg = ""
    for i, arg_info in enumerate(arg_infos):
        args_lists = arg_info["args"]
        kind_arg = arg_info.get("kind_arg", False)
        same_kind_arg = arg_info.get("same_kind_arg", False)
        no_of_args = len(args_lists[0])
        no_of_args_msg += " or " if i > 0 else ""
        no_of_args_msg += f"{no_of_args + int(kind_arg)}"
        else_if = "else if" if i > 0 else "if"
        src += 2 * indent + f"{else_if} (args.size() == {no_of_args + int(kind_arg)}) " + " {\n"
        compute_arg_types(3 * indent, no_of_args, "args")
        allow_polymorphic_arg = arg_info.get("allow_polymorphic_arg", None)
        condition, cond_in_msg = compute_arg_condition(no_of_args, args_lists, allow_polymorphic_arg, "args")
        src += 3 * indent + f'if(!({condition}))' + ' {\n'
        src += 4 * indent + f'append_error(diag, "Unexpected args, {func_name} expects {cond_in_msg} as arguments", loc);\n'
        src += 4 * indent + f'return nullptr;\n'
        src += 3 * indent + '}\n'
        if same_kind_arg:
            compute_arg_kinds(3 * indent, same_kind_arg)
            condition = compute_kind_condition(same_kind_arg)
            src += 3 * indent + f'if(!({condition}))' + ' {\n'
            src += 4 * indent + f'append_error(diag, "Kind of all the arguments of {func_name} must be the same", loc);\n'
            src += 4 * indent + f'return nullptr;\n'
            src += 3 * indent + '}\n'
        kind_validation_info = arg_info.get("kind_validation", [])
        if kind_validation_info != []:
            src += 3 * indent + "int kind = 0;\n"
        for validation_item in kind_validation_info:
            for arg_name, arg_spec in validation_item.items():
                arg_pos = list(arg_spec.keys())[0]
                required_kind = list(arg_spec.values())[0]
                src += 3 * indent + f"kind = ASRUtils::extract_kind_from_ttype_t(expr_type(args[{arg_pos}]));\n"
                src += 3 * indent + f"if(kind != {required_kind}) " + "{\n"
                src += 4 * indent + f'append_error(diag, "{arg_name} argument of `{func_name.lower()}` must have kind equal to {required_kind}", loc);\n'
                src += 4 * indent + f'return nullptr;\n'
                src += 3 * indent + '}\n'
        char_len_validation_info = arg_info.get("char_len_validation", [])
        for validation_item in char_len_validation_info:
            for arg_name, arg_spec in validation_item.items():
                arg_pos = list(arg_spec.keys())[0]
                constraint = list(arg_spec.values())[0]
                src += 3 * indent + f'if (ASR::is_a<ASR::String_t>(*arg_type{arg_pos})) {{\n'
                src += 4 * indent + f'int64_t len = ASRUtils::get_fixed_string_len(arg_type{arg_pos});\n'
                # If type-based length is unknown, try to get length from compile-time value
                src += 4 * indent + f'if (len == -1) {{\n'
                src += 5 * indent + f'ASR::expr_t* value = ASRUtils::expr_value(args[{arg_pos}]);\n'
                src += 5 * indent + f'if (value && ASR::is_a<ASR::StringConstant_t>(*value)) {{\n'
                src += 6 * indent + f'len = std::string(ASR::down_cast<ASR::StringConstant_t>(value)->m_s).size();\n'
                src += 5 * indent + f'}}\n'
                src += 4 * indent + f'}}\n'
                if constraint["min"] == constraint["max"]:
                    src += 4 * indent + f'if (len != -1 && len != {constraint["min"]}) {{\n'
                    src += 5 * indent + f'append_error(diag, "Argument {arg_name} to {func_name} must have length {constraint["min"]}", loc);\n'
                else:
                    src += 4 * indent + f'if (len != -1 && (len < {constraint["min"]} || len > {constraint["max"]})) {{\n'
                    src += 5 * indent + f'append_error(diag, "Argument {arg_name} to {func_name} must have length between {constraint["min"]} and {constraint["max"]}", loc);\n'
                src += 5 * indent + 'return nullptr;\n'
                src += 4 * indent + '}\n'
                src += 3 * indent + '}\n'

        src += 2 * indent + "}\n"
    src += 2 * indent + "else {\n"
    src += 3 * indent + f'append_error(diag, "Unexpected number of args, {func_name} takes {no_of_args_msg} arguments, found " + std::to_string(args.size()), loc);\n'
    src += 3 * indent + f'return nullptr;\n'
    src += 2 * indent + "}\n"


def add_create_func_return_src(func_name):
    global src, indent
    arg_infos = intrinsic_funcs_args[func_name]
    args_lists = arg_infos[0]["args"]
    no_of_args = len(args_lists[0])
    ret_type_val = arg_infos[0].get("return", None)
    ret_type_arg_idx = arg_infos[0].get("ret_type_arg_idx", None)
    if ret_type_val:
        ret_type = ret_type_val
    else:
        src += indent * 2 + "ASRUtils::ExprStmtDuplicator expr_duplicator(al);\n"
        src += indent * 2 + "expr_duplicator.allow_procedure_calls = true;\n"
        if ( ret_type_arg_idx == "dynamic"):
            src += indent * 2 + f"int upper_kind = 0;\n"
            src += indent * 2 + f"for(size_t i=0;i<args.size();i++){{\n"
            src += indent * 3 + f"upper_kind = std::max(upper_kind,ASRUtils::extract_kind_from_ttype_t(expr_type(args[i])));\n"
            src += indent * 2 + f"}}\n"
            src += indent * 2 + f"ASR::ttype_t* type_ = expr_duplicator.duplicate_ttype(ASRUtils::extract_type(expr_type(args[0])));\n"
            src += indent * 2 + f"set_kind_to_ttype_t(type_,upper_kind);\n"
        else:
            src += indent * 2 + "ASR::ttype_t* type_ = nullptr;\n"
            src += indent * 2 + f"type_ = expr_duplicator.duplicate_ttype(ASRUtils::extract_type(expr_type(args[{ret_type_arg_idx}])));\n"
        ret_type = "type_"
    kind_arg = arg_infos[0].get("kind_arg", False)
    src += indent * 2 + f"ASR::ttype_t *return_type = {ret_type};\n"
    if kind_arg:
        src += indent * 2 + "if ( args[1] != nullptr ) {\n"
        src += indent * 3 +     "int kind = -1;\n"
        src += indent * 3 +     "if (!ASR::is_a<ASR::Integer_t>(*expr_type(args[1])) || !extract_value(ASRUtils::expr_value(args[1]), kind)) {\n"
        src += indent * 4 +         f'append_error(diag, "`kind` argument of the `{func_name}` function must be a scalar Integer constant", args[1]->base.loc);\n'
        src += indent * 4 +         "return nullptr;\n"
        src += indent * 3 +     "}\n"
        src += indent * 3 +     "set_kind_to_ttype_t(return_type, kind);\n"
        src += indent * 2 + "}\n"
    real_32_except_complex = arg_infos[0].get("real_32_except_complex", False)
    if real_32_except_complex:
        src += indent * 2 + "else { \n"
        src += indent * 3 + "ASR::ttype_t* arg_type = ASRUtils::expr_type(args[0]);\n"
        src += indent * 3 + "if (is_complex(*arg_type)) { \n"
        src += indent * 4 + "int kind = ASRUtils::extract_kind_from_ttype_t(arg_type); \n"
        src += indent * 4 + "set_kind_to_ttype_t(return_type, kind); \n"
        src += indent * 3 + "} \n"
        src += indent * 2 + "} \n"
    src += indent * 2 + "ASR::expr_t *m_value = nullptr;\n"
    src += indent * 2 + f"Vec<ASR::expr_t*> m_args; m_args.reserve(al, {no_of_args});\n"
    for _i in range(no_of_args):
        src += indent * 2 + f"m_args.push_back(al, args[{_i}]);\n"
    if func_name in compile_time_only_fn:
        src += indent * 2 + f"return_type = ASRUtils::extract_type(return_type);\n"
        src += indent * 2 + f"m_value = eval_{func_name}(al, loc, return_type, args, diag);\n"
        src += indent * 3 +     "if (diag.has_error()) {\n"
        src += indent * 4 +         f"return nullptr;\n"
        src += indent * 3 +     "}\n"
        src += indent * 2 + "if (m_value) {\n"
        src += indent * 3 + "return ASR::make_TypeInquiry_t(al, loc, "\
            f"static_cast<int64_t>(IntrinsicElementalFunctions::{func_name}), "\
            "ASRUtils::expr_type(m_args[0]), m_args[0], return_type, m_value);\n"
        src += indent * 2 + "}\n"
        src += indent * 2 + f"return ASR::make_IntrinsicElementalFunction_t(al, loc, "\
            f"static_cast<int64_t>(IntrinsicElementalFunctions::{func_name}), "\
            "m_args.p, m_args.n, 0, return_type, m_value);\n"

    else:
        src += indent * 2 +     f"for( size_t i = 0; i < {no_of_args}; i++ ) " + "{\n"
        src += indent * 3 +         "ASR::ttype_t* type = ASRUtils::expr_type(args[i]);\n"
        src += indent * 3 +         "if (ASRUtils::is_array(type)) {\n"
        src += indent * 4 +             "ASR::dimension_t* m_dims = nullptr;\n"
        src += indent * 4 +             "size_t n_dims = ASRUtils::extract_dimensions_from_ttype(type, m_dims);\n"
        src += indent * 4 +             "return_type = ASRUtils::make_Array_t_util(al, type->base.loc, "
        src +=                              "return_type, m_dims, n_dims, ASR::abiType::Source, false, "
        src +=                              "ASR::array_physical_typeType::DescriptorArray);\n"
        src += indent * 4 +             "break;\n"
        src += indent * 3 +         "}\n"
        src += indent * 2 +     "}\n"

        src += indent * 2 + "if (all_args_evaluated(m_args)) {\n"
        src += indent * 3 +     f"Vec<ASR::expr_t*> args_values; args_values.reserve(al, {no_of_args});\n"
        for _i in range(no_of_args):
            src += indent * 3 + f"args_values.push_back(al, expr_value(m_args[{_i}]));\n"
        src += indent * 3 +     f"m_value = eval_{func_name}(al, loc, return_type, args_values, diag);\n"
        src += indent * 3 +     "if (diag.has_error()) {\n"
        src += indent * 4 +         f"return nullptr;\n"
        src += indent * 3 +     "}\n"
        src += indent * 2 + "}\n"
        if "null" in intrinsic_funcs_ret_type.get(func_name, []):
            src += indent * 2 + f"return ASR::make_Expr_t(al, loc, ASRUtils::EXPR(ASR::make_IntrinsicElementalFunction_t(al, loc, static_cast<int64_t>(IntrinsicElementalFunctions::{func_name}), m_args.p, m_args.n, 0, return_type, m_value)));\n"
        else:
            src += indent * 2 + f"return ASR::make_IntrinsicElementalFunction_t(al, loc, static_cast<int64_t>(IntrinsicElementalFunctions::{func_name}), m_args.p, m_args.n, 0, return_type, m_value);\n"

def gen_verify_args(func_name):
    global src
    src += indent + R"static inline void verify_args(const ASR::IntrinsicElementalFunction_t& x, diag::Diagnostics& diagnostics) {" + "\n"
    add_verify_arg_type_src(func_name)
    runtime_fallback_fns = ["SameTypeAs", "ExtendsTypeOf"]
    if func_name in compile_time_only_fn and func_name not in runtime_fallback_fns:
        src += indent * 2 + 'ASRUtils::require_impl(x.m_value, '\
            f'"Missing compile time value, `{func_name}` intrinsic output must '\
            'be computed during compile time", x.base.base.loc, diagnostics);\n'
    add_verify_return_type_src(func_name)
    src += indent + "}\n\n"

def gen_create_function(func_name):
    global src
    src += indent + Rf"static inline ASR::asr_t* create_{func_name}(Allocator& al, const Location& loc, Vec<ASR::expr_t*>& args, diag::Diagnostics& diag) " + "{\n"
    add_create_func_arg_type_src(func_name)
    add_create_func_return_src(func_name)
    src += indent + "}\n"


def get_registry_funcs_src():
    global src
    for func_name in intrinsic_funcs_args.keys():
        src += f"namespace {func_name}" + " {\n\n"
        gen_verify_args(func_name)

        if func_name not in skip_create_func:
            gen_create_function(func_name)
        src += "}\n\n"
    return src


HEAD = """// This file is AUTO-GENERATED by src/libasr/intrinsic_func_registry_util_gen.py
// Do not edit this file directly. Instead, modify intrinsic_func_registry_util_gen.py
// and regenerate this file by running:
//     python src/libasr/intrinsic_func_registry_util_gen.py

#ifndef LIBASR_PASS_INTRINSIC_FUNC_REG_UTIL_H
#define LIBASR_PASS_INTRINSIC_FUNC_REG_UTIL_H

#include <libasr/asr_utils.h>
#include <libasr/pass/intrinsic_functions.h>

namespace LCompilers {

namespace ASRUtils {

"""

FOOT = """
} // namespace ASRUtil

} // namespace LCompilers

#endif // LIBASR_PASS_INTRINSIC_FUNC_REG_UTIL_H
"""
def main(argv):
    if len(argv) == 2:
        out_file = argv[1]
    elif len(argv) == 1:
        print("Assuming default values of intrinsic_function_registry_util.h")
        here = os.path.dirname(__file__)
        pass_dir = os.path.join(here, "pass")
        out_file = os.path.join(pass_dir, "intrinsic_function_registry_util.h")
    else:
        print("invalid arguments")
        return 2
    fp = open(out_file, "w", encoding="utf-8")
    try:
        fp.write(HEAD)
        fp.write(get_registry_funcs_src())
        fp.write(FOOT)
    finally:
        fp.close()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
# Design

This file converts from a GFortran module file representation (documented in
the `lfortran.adapters.gfortran.mod` module) to an Abstract Semantic Representation (ASR).
"""

# TODO: move this into the lfortran package itself
import sys
sys.path.append("../..")

from lfortran.asr import asr
from lfortran.asr.asr_check import verify_asr
from lfortran.asr.asr_to_ast import asr_to_ast
from lfortran.ast.ast_to_src import ast_to_src
from lfortran.adapters.gfortran.mod import mod_to_asr
from lfortran.asr.builder import (make_translation_unit,
    translation_unit_make_module, scope_add_function, make_type_integer,
    make_type_real, type_eq, make_binop, scope_add_symbol, Scope,
    function_make_var, array_is_assumed_shape)

class WrapperVisitor(asr.ASTVisitor):

    def visit_TranslationUnit(self, node):
        out = ""
        for s in node.global_scope.symbols:
            sym = node.global_scope.symbols[s]
            out += self.visit(sym)
        return out

    def visit_Module(self, node):
        name = "mod2"
        out = """\
#ifndef MOD2_H
#define MOD2_H

extern "C" {

"""
        out += self.visit_scope(node.symtab)
        out += """\
}

#endif // MOD2_H
"""
        return out

    def visit_scope(self, symtab, parent=None):
        out = ""
        for s, sym in symtab.symbols.items():
            out += self.visit(sym)
        return out

    def visit_object(self, o):
        if isinstance(o, Scope):
            return self.visit_scope(o)
        elif isinstance(o, list):
            return self.visit_sequence(o)
        elif isinstance(o, (str, int)) or o is None:
            return o
        else:
            print(type(o))
            raise NotImplementedError()

    def visit_Function(self, node):
        name = self.visit_object(node.name)
        args2 = []
        for arg in node.args:
            #type = self.visit(arg.type)
            #if array_is_assumed_shape(type):
            #    if len(type.dims) == 1:
            #        dname = "c_desc1_t"
            #    elif len(type.dims) == 2:
            #        dname = "c_desc2_t"
            #    else:
            #        raise NotImplementedError("Too many dimensions")
            #    args2.append(dname)
            #else:
            #    args2.append(arg)
            n = len(arg.type.dims)
            if n == 0:
                args2.append("int32_t *%s" % arg.name)
            else:
                if array_is_assumed_shape(arg.type):
                    args2.append("descriptor<%s, int32_t> *%s" % (n, arg.name))
                else:
                    args2.append("int32_t *%s" % arg.name)
        #symtab = self.visit_object(node.symtab)
        #self._lookup = 1
        #self._scope = symtab
        #args = self.visit_sequence(node.args)
        #body = self.visit_sequence(node.body)
        #return_var = self.visit(node.return_var)
        #self._lookup = 0
        #if node.bind:
        #    bind = self.visit(node.bind)
        #else:
        #    bind = None
        return "int32_t __mod1_MOD_%s(%s);\n" % (name, ", ".join(args2))
        #return asr.Function(name=name, args=args, body=body, bind=bind, return_var=return_var, symtab=symtab)



def create_wrapper(u):
    v = WrapperVisitor()
    u2 = v.visit(u)
    return u2

u = mod_to_asr("mod1.mod")
u2 = create_wrapper(u)
print(u2)



import subprocess

# Open file_path, Read it with mode 'r' which expects a utf-8 encode.
# If the file is binary it'd conflict with
# text mode 'r', leading to raised error `UnicodeDecodeError`.
def is_file_binary(file_path: str) -> bool:
    try:
        with open(file_path, 'r') as fp:
            fp.read(16)
            return False
    except UnicodeDecodeError:
            return True


# Get all commit hashes difference between `origin/main` and current `HEAD`
def get_commit_hash_differences() -> list :
    # Run command `git rev-list origin/main..HEAD`
    # Use `subporcess.PIPE` as the output buffer to fetch the list of commit hashes
    completedProcess_object =  subprocess.run(["git" ,"rev-list" ,"origin/main..HEAD"],
                                                stdout=subprocess.PIPE
                                                , text=True)
    # Create list out of the resulting commit hases. 
    commit_hashes_listed  : list  = completedProcess_object.stdout.split("\n")
    if(commit_hashes_listed[-1] == ""):
        commit_hashes_listed.remove("") # Remove trailing empty string.
    print("number of commits from current PR branch against origin/main is :", len(commit_hashes_listed))
    print (completedProcess_object.stdout) # print the whole string of commit hashes.
    return commit_hashes_listed

# Fetch all files pathes that has beend changed from current `HEAD` against `origin/main`.
def get_changed_files() -> list :
    # Run command `git diff --name-only origin/main --diff-filter=d`
    # Notice we filter out the deleted files.
    completedProcess_object = subprocess.run(["git","diff","--name-only","origin/main", "--diff-filter=d"],
                                            stdout=subprocess.PIPE, text=True)
    file_paths : list = completedProcess_object.stdout.split("\n")
    if(file_paths[-1] == ""): file_paths.remove("") # Remove trailing empty string.
    return file_paths




def run_test() :
    commit_hashes_list = get_commit_hash_differences()
    # Run git `checkout` to each commit_hash between `origin/main` --> `HEAD`
    for commit_hash in commit_hashes_list:
        subprocess.run(["git","checkout",commit_hash], 
                       stdout=subprocess.PIPE,stderr=subprocess.PIPE,text=True) 
        print("checked out to commit --> ",commit_hash)

        file_paths : list = get_changed_files() # fetch changed files from current-checkout-to commit.
        for file_path in file_paths:
            if (is_file_binary(file_path)):
                print("FAIL --- binary file detected")
                print(f"Binary File Name : ({file_path}) In Commit : ({commit_hash})")
                raise SystemExit(1)

        print("OK --- no binary file found")
        print("\n--------------------------------------------------\n")


    
    print("CLEAN : No binary file found in any commit in this PR")
    print("\n--------------------------------------------------\n")
    if(len(commit_hashes_list) > 0):
        print("==> Checking out to latest commit in the PR branch")
        print("==> Commit hash :", commit_hashes_list[0])
        subprocess.run(["git","checkout",commit_hashes_list[0]], 
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,text=True)
        print("\n--------------------------------------------------\n")
    print("DONE")


run_test()




#!/usr/bin/env python3

from struct import unpack
from sys import argv
from re import sub

lines = ""
with open(argv[1], "rb") as f:
    lines = f.read()

list = []
for i in range(0, len(lines), 24):
    list.append(sub('[(),]', '', str(unpack("3Q", lines[i:i+24]))))

with open(argv[1] + ".txt", "w") as f:
    j = 0
    for i in list:
        f.write(i+'\n')



import os
import shutil
import sys
from pathlib import Path
from typing import Iterator, Optional

import pytest

from lfortran_language_server.lfortran_lsp_test_client import LFortranLspTestClient


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--execution-strategy",
        action="store",
        default="concurrent",
        help="Specifies the execution strategy for handling messages. The `parallel` strategy implies multiple messages may be processed alongside each other, while the `concurrent` strategy implies multiple messages may be processed but only one processor will be active at a time (they will yield control to each other).",
        choices=("concurrent", "parallel"),
    )


@pytest.fixture
def client(request: pytest.FixtureRequest, capfd: pytest.CaptureFixture) -> Iterator[LFortranLspTestClient]:
    execution_strategy = request.config.getoption("--execution-strategy")

    server_path = None
    if 'LFORTRAN_PATH' in os.environ:
        server_path = os.environ['LFORTRAN_PATH']
    if server_path is None or not os.path.exists(server_path):
        server_path = Path(__file__).absolute().parent.parent.parent.parent / "src" / "bin" / "lfortran"
    if server_path is None or not os.path.exists(server_path):
        server_path = shutil.which('lfortran')
    if server_path is None:
        raise RuntimeError('cannot determine location of lfortran')
    server_path = Path(server_path)
    if not (server_path.exists() and os.access(server_path, os.X_OK)):
        raise RuntimeError(f'Invalid or non-executable path to lfortran: {server_path}')

    compiler_path = server_path

    server_log_path = f"{request.node.name}-{execution_strategy}-server.log"
    client_log_path = f"{request.node.name}-{execution_strategy}-client.log"
    stdout_log_path = f"{request.node.name}-{execution_strategy}-stdout.log"
    stdin_log_path = f"{request.node.name}-{execution_strategy}-stdin.log"
    gdb_log_path = f"{request.node.name}-{execution_strategy}-gdb.log"
    lldb_log_path = f"{request.node.name}-{execution_strategy}-lldb.log"

    config = {
        "LFortran": {
            "openIssueReporterOnError": False,
            "maxNumberOfProblems": 100,
            "trace": {
                "server": "verbose",
            },
            "compiler": {
                "path": "lfortran",
                "flags": [],
            },
            "log": {
                "path": server_log_path,
                "level": "all",
                "prettyPrint": False,
            },
            "indentSize": 4,
            # NOTE: Regarding timing-out and retrying requests, a timeout of 0
            # means do not time-out (or retry):
            "timeoutMs": 0,
            "retry": {
                "maxAttempts": 3,
                "minSleepTimeMs": 10,
                "maxSleepTimeMs": 300,
            },
            "telemetry": {
                "enabled": True,
                "frequencyMs": 1000,
            },
        }
    }

    server_args = []

    # FIXME: Find the correct combination of commands to get LLDB to dump crashes
    # to a log file like GDB and uncomment the following to enable it:
    # ---------------------------------------------------------------------------
    # lldb_path = shutil.which('lldb')
    lldb_path = None  #<- FIXME: When LLDB is enabled, remove this line.

    # NOTE: If you have GDB on your system and would like to run the tests with
    # it, uncomment the following line so GDB may be found and disable the line
    # that assigns `None` to it:
    # -------------------------------------------------------------------------
    # gdb_path = shutil.which('gdb')
    gdb_path = None

    if lldb_path is not None:
        server_args += [
            "-q",
            "-b",
            "-o", "run",
            "-o", "bt",
            "-o", f"log enable lldb crashlog {lldb_log_path}",
            "-o", "quit",
            str(compiler_path),
            "--",
        ]
        server_path = Path(lldb_path)
    elif gdb_path is not None:
        server_args += [
            "-q", "-batch",
            "-ex", "set logging enabled off",
            "-ex", "set logging redirect on",
            "-ex", "set logging debugredirect on",
            "-ex", "set logging overwrite on",
            "-ex", f"set logging file {gdb_log_path}",
            "-ex", "set logging enabled on",
            "-ex", "run",
            "-ex", "bt",
            "--args", str(compiler_path),
        ]
        server_path = Path(gdb_path)

    server_args += [
        "server",
        "--parent-process-id", str(os.getpid()),
        "--log-level", config["LFortran"]["log"]["level"],
        "--log-path", server_log_path,
        "--timeout-ms", str(config["LFortran"]["timeoutMs"]),
        "--num-request-threads", "1",
        "--num-worker-threads", "1",
        "--config-section", "LFortran",
        "--open-issue-reporter-on-error", str(config["LFortran"]["openIssueReporterOnError"]).lower(),
        "--max-number-of-problems", str(config["LFortran"]["maxNumberOfProblems"]),
        "--trace-server", config["LFortran"]["trace"]["server"],
        "--compiler-path", str(compiler_path),
        "--log-pretty-print", str(config["LFortran"]["log"]["prettyPrint"]).lower(),
        "--indent-size", str(config["LFortran"]["indentSize"]),
        "--max-retry-attempts", str(config["LFortran"]["retry"]["maxAttempts"]),
        "--min-retry-sleep-time-ms", str(config["LFortran"]["retry"]["minSleepTimeMs"]),
        "--max-retry-sleep-time-ms", str(config["LFortran"]["retry"]["maxSleepTimeMs"]),
        "--extension-id", "lcompilers.lfortran",
        "--execution-strategy", execution_strategy,
    ]

    def print_log(log_path: str, heading: str) -> None:
        header = f"~~ {heading} [{log_path}] ~~"
        border = "~" * len(header)
        print(file=sys.stderr)
        print(border, file=sys.stderr)
        print(header, file=sys.stderr)
        print(border, file=sys.stderr)
        print(file=sys.stderr)
        if os.path.exists(log_path):
            with open(log_path) as f:
                print(f.read(), file=sys.stderr)
        else:
            print(f"Log file does not exist: {log_path}", file=sys.stderr)

    def print_logs() -> None:
        print_log(client_log_path, "Client Logs")
        print_log(stdin_log_path, "Standard Input")
        print_log(stdout_log_path, "Standard Output")
        print_log(server_log_path, "Server Logs")
        if lldb_path is  not None:
            print_log(lldb_log_path, "Low-Level Debugger (LLDB)")
        elif gdb_path is not None:
            print_log(gdb_log_path, "GNU Debugger (GDB)")

    client: Optional[LFortranLspTestClient] = None
    logs_printed = False
    try:
        client = LFortranLspTestClient(
            server_path=server_path,
            server_params=server_args,
            workspace_path=None,
            timeout_ms=3000,
            config=config,
            client_log_path=client_log_path,
            stdout_log_path=stdout_log_path,
            stdin_log_path=stdin_log_path
        )

        with client.serve():
            # Steps abstracted by context manager:
            # 1. Send request: initialize
            # 2. Send notification: initialized
            # 3. <yield client>
            # 4. Send request: shutdown
            # 5. Send notification: exit
            yield client
    except BaseException as e:
        if not isinstance(e, SystemExit) or e.code != 0:
            print_logs()
            logs_printed = True
        raise e
    finally:
        try:
            if client is not None and hasattr(client, 'server') \
               and client.server.poll():
                client.kill_server()
        finally:
            if not logs_printed:
                print_logs()

    # Consuming stderr should prevent leaks
    capfd.readouterr()
